# automatically generated by the FlatBuffers compiler, do not modify

# namespace: Dl

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class StringStringEntry(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = StringStringEntry()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsStringStringEntry(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # StringStringEntry
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # StringStringEntry
    def Key(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # StringStringEntry
    def Value(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

def StringStringEntryStart(builder):
    builder.StartObject(2)

def Start(builder):
    StringStringEntryStart(builder)

def StringStringEntryAddKey(builder, key):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(key), 0)

def AddKey(builder, key):
    StringStringEntryAddKey(builder, key)

def StringStringEntryAddValue(builder, value):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(value), 0)

def AddValue(builder, value):
    StringStringEntryAddValue(builder, value)

def StringStringEntryEnd(builder):
    return builder.EndObject()

def End(builder):
    return StringStringEntryEnd(builder)


class StringStringEntryT(object):

    # StringStringEntryT
    def __init__(self):
        self.key = None  # type: str
        self.value = None  # type: str

    @classmethod
    def InitFromBuf(cls, buf, pos):
        stringStringEntry = StringStringEntry()
        stringStringEntry.Init(buf, pos)
        return cls.InitFromObj(stringStringEntry)

    @classmethod
    def InitFromPackedBuf(cls, buf, pos=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, pos)
        return cls.InitFromBuf(buf, pos+n)

    @classmethod
    def InitFromObj(cls, stringStringEntry):
        x = StringStringEntryT()
        x._UnPack(stringStringEntry)
        return x

    # StringStringEntryT
    def _UnPack(self, stringStringEntry):
        if stringStringEntry is None:
            return
        self.key = stringStringEntry.Key()
        self.value = stringStringEntry.Value()

    # StringStringEntryT
    def Pack(self, builder):
        if self.key is not None:
            key = builder.CreateString(self.key)
        if self.value is not None:
            value = builder.CreateString(self.value)
        StringStringEntryStart(builder)
        if self.key is not None:
            StringStringEntryAddKey(builder, key)
        if self.value is not None:
            StringStringEntryAddValue(builder, value)
        stringStringEntry = StringStringEntryEnd(builder)
        return stringStringEntry
